#define BUILD_NUMBER \
	"build-14549434"
#define BUILD_NUMBER_NUMERIC \
	14549434
#define BUILD_NUMBER_NUMERIC_STRING \
	"14549434"
#define PRODUCT_BUILD_NUMBER \
	"product-build-15108"
#define PRODUCT_BUILD_NUMBER_NUMERIC \
	15108
#define PRODUCT_BUILD_NUMBER_NUMERIC_STRING \
	"15108"
